==> dataiku-dss-14.2.1.tar.gz: dataiku-dss-14.2.1/python/dataiku/doctor/preprocessing/generated_features_mapping.py <==
class GeneratedFeaturesMapping:
    # One-to-many mapping
    SINGLE_ORIGIN_COLUMN = "single_origin_column"
    # One-to-one mapping
    SINGLE_PROCESSED_FEATURE = "single_processed_feature"

    def __setstate__(self, state):
        """
        We have this for backwards compatibility reasons. Ensemble models are pickled with their pipelines,
        and before version 12.4 models didn't have the "feature_to_block" property.
        Therefore, once they are unserialized,  they are missing the "feature_to_block" property.
        We create it during the unpickling so that the generated_features_mapping
        is populated during `pipeline.process`.
        """
        self.__dict__.update(state)
        if not hasattr(self, "feature_to_block"):
            self.feature_to_block = {}

    def __init__(self):
        self.mapping = {}
        self.feature_to_block = {}

    def add_whole_block_mapping(self, block_name, original_names):
        self.mapping[block_name] = {
            "type": self.SINGLE_ORIGIN_COLUMN,
            "original_names": original_names,
        }

    def add_per_column_mapping(self, block_name, original_names, new_name):
        if block_name not in self.mapping:
            block_type = self.SINGLE_PROCESSED_FEATURE
            self.mapping[block_name] = {"type": block_type, "values": {}, }
        self.feature_to_block[new_name] = block_name
        self.mapping[block_name]["values"][new_name] = original_names

    def add_features_to_block(self, features, block_name):
        """
        Map all features to their original block name
        :param features: list of features
        :param block_name: name of the block containing features
        """
        if features is None:
            return
        for feature in features:
            self.feature_to_block[feature] = block_name

    def get_block_from_feature(self, feature_name):
        """
        This method has 2 distinct logics.
        For sparse features whose block is easy to track back we infer the block.
        For other features we use a simple dictionary
        :param feature_name: str
        :return: name of the block
        """
        feature_generation = feature_name.split(":")[0]
        # The safe preprocessing are because the last part is an idx and therefore not a user inputted value that can contain ":"
        # We do this for blocks that can have a large number of features
        if feature_generation in ["hashing", "unfold", "hashvect"]:
            return ":".join(feature_name.split(":")[:-1])
        return self.feature_to_block[feature_name]

    def should_map_whole_block_to_one_origin_column(self, block_name):
        if block_name not in self.mapping:
            return False
        else:
            return self.mapping[block_name]["type"] == self.SINGLE_ORIGIN_COLUMN

    def get_whole_block_original(self, block_name):
        return self.mapping[block_name]["original_names"]

    def get_origin_columns_from_mapping(self, block_name, feature_name):
        block_mapping = self.mapping[block_name]
        if block_mapping.get("type") == self.SINGLE_ORIGIN_COLUMN:
            return block_mapping["original_names"]
        else: # SINGLE_PROCESSED_FEATURE
            return block_mapping["values"][feature_name]

    def get_origin_columns_from_feature(self, feature_name, columns):
        # Feature wasn't generated
        if feature_name in columns:
            return [feature_name]

        block_name = self.get_block_from_feature(feature_name)
        if block_name in self.mapping:
            return self.get_origin_columns_from_mapping(block_name, feature_name)
        raise ValueError("Unsupported feature mapping {}".format(feature_name))

    def get_features_from_original_column(self, original_column):
        features = []
        for block_name, block in self.mapping.items():
            if block.get("type") == self.SINGLE_ORIGIN_COLUMN:
                # one-to-many mapping
                if original_column in block["original_names"]:
                    features.extend(self.get_features_from_block_name(block_name))
            else:
                # SINGLE_PROCESSED_FEATURE one-to-one mapping
                for feature, original_names in block["values"].items():
                    if original_column in original_names:
                        features.append(feature)
        return features

    def get_features_from_block_name(self, block_name):
        features = []
        if block_name is not None:
            for feature, current_block_name in self.feature_to_block.items():
                if current_block_name == block_name:
                    features.append(feature)
        return features

==> dataiku-dss-14.3.0.tar.gz: dataiku-dss-14.3.0/python/dataiku/doctor/preprocessing/generated_features_mapping.py <==
class GeneratedFeaturesMapping:
    # One-to-many mapping
    SINGLE_ORIGIN_COLUMN = "single_origin_column"
    # One-to-one mapping
    SINGLE_PROCESSED_FEATURE = "single_processed_feature"

    def __setstate__(self, state):
        """
        We have this for backwards compatibility reasons. Ensemble models are pickled with their pipelines,
        and before version 12.4 models didn't have the "feature_to_block" property.
        Therefore, once they are unserialized,  they are missing the "feature_to_block" property.
        We create it during the unpickling so that the generated_features_mapping
        is populated during `pipeline.process`.
        """
        self.__dict__.update(state)
        if not hasattr(self, "feature_to_block"):
            self.feature_to_block = {}

    def __init__(self):
        self.mapping = {}
        self.feature_to_block = {}

    def add_whole_block_mapping(self, block_name, original_names):
        self.mapping[block_name] = {
            "type": self.SINGLE_ORIGIN_COLUMN,
            "original_names": original_names,
        }

    def add_per_column_mapping(self, block_name, original_names, new_name):
        if block_name not in self.mapping:
            block_type = self.SINGLE_PROCESSED_FEATURE
            self.mapping[block_name] = {"type": block_type, "values": {}, }
        self.feature_to_block[new_name] = block_name
        self.mapping[block_name]["values"][new_name] = original_names

    def add_features_to_block(self, features, block_name):
        """
        Map all features to their original block name
        :param features: list of features
        :param block_name: name of the block containing features
        """
        if features is None:
            return
        for feature in features:
            self.feature_to_block[feature] = block_name

    def get_block_from_feature(self, feature_name):
        """
        This method has 2 distinct logics.
        For sparse features whose block is easy to track back we infer the block.
        For other features we use a simple dictionary
        :param feature_name: str, name of the preprocessed feature
        :return: name of the block containing the feature
        """
        feature_generation = feature_name.split(":")[0]
        # The safe preprocessing are because the last part is an idx and therefore not a user inputted value that can contain ":"
        # We do this for blocks that can have a large number of features
        if feature_generation in ["hashing", "unfold", "hashvect"]:
            return ":".join(feature_name.split(":")[:-1])
        return self.feature_to_block[feature_name]

    def should_map_whole_block_to_one_origin_column(self, block_name):
        if block_name not in self.mapping:
            return False
        else:
            return self.mapping[block_name]["type"] == self.SINGLE_ORIGIN_COLUMN

    def get_whole_block_original(self, block_name):
        return self.mapping[block_name]["original_names"]

    def get_origin_columns_from_mapping(self, block_name, feature_name):
        block_mapping = self.mapping[block_name]
        if block_mapping.get("type") == self.SINGLE_ORIGIN_COLUMN:
            return block_mapping["original_names"]
        else: # SINGLE_PROCESSED_FEATURE
            return block_mapping["values"][feature_name]

    def get_origin_columns_from_feature(self, feature_name, origin_columns):
        """
        :param feature_name: str, name of preprocessed feature
        :param origin_columns: list of str, names of input columns (from the dataset)
        :return: list of str, names of input columns (from the dataset) that are preprocessed into `feature_name`
        """
        # Feature wasn't generated
        if feature_name in origin_columns:
            return [feature_name]

        block_name = self.get_block_from_feature(feature_name)
        if block_name in self.mapping:
            return self.get_origin_columns_from_mapping(block_name, feature_name)
        raise ValueError("Unsupported feature mapping {}".format(feature_name))

    def get_features_from_origin_column(self, origin_column):
        """
        :param origin_column: str, name of input column (from the dataset)
        :return: list of str, names of the preprocessed features that are based on `origin_column`, excluding rolling windows
        """
        features = []
        for block_name, block in self.mapping.items():
            if block_name == "rolling_window":
                continue
            if block.get("type") == self.SINGLE_ORIGIN_COLUMN:
                # one-to-many mapping
                if origin_column in block["original_names"]:
                    features.extend(self.get_features_from_block_name(block_name))
            else:
                # SINGLE_PROCESSED_FEATURE one-to-one mapping
                for feature, original_names in block["values"].items():
                    if origin_column in original_names:
                        features.append(feature)
        return features

    def get_features_from_block_name(self, block_name):
        features = []
        if block_name is not None:
            for feature, current_block_name in self.feature_to_block.items():
                if current_block_name == block_name:
                    features.append(feature)
        return features
